import hashlib
import hmac
import json
from datetime import datetime

from django.contrib.auth.models import Group
from django.core import mail
from django.test import TestCase
from django.utils import timezone

from job_runner.apps.job_runner.models import (
    Job,
    JobTemplate,
    KillRequest,
    Project,
    Run,
    RunLog,
    Worker,
)


class ApiTestBase(TestCase):
    """
    Base class for API testing.
    """
    def get(self, path, *args, **kwargs):
        api_key = hmac.new(
            'verysecret', 'GET{0}'.format(path), hashlib.sha1).hexdigest()

        return self.client.get(
            path,
            *args,
            ACCEPT='application/json',
            HTTP_AUTHORIZATION='ApiKey worker1:{0}'.format(api_key)
        )

    def get_json(self, path, *args, **kwargs):
        response = self.get(path, *args, **kwargs)
        return json.loads(response.content)

    def post(self, path, data, method_override='POST'):
        json_data = json.dumps(data)
        api_key = hmac.new(
            'verysecret', '{0}{1}{2}'.format(method_override, path, json_data),
            hashlib.sha1
        ).hexdigest()

        return self.client.post(
            path,
            data=json_data,
            content_type='application/json',
            ACCEPT='application/json',
            HTTP_AUTHORIZATION='ApiKey worker1:{0}'.format(api_key),
            HTTP_X_HTTP_METHOD_OVERRIDE=method_override,
        )

    def patch(self, path, data):
        return self.post(path, data, method_override='PATCH')


class ProjectTestCase(ApiTestBase):
    """
    Tests for the project interface.
    """
    fixtures = ['test_auth', 'test_project', 'test_worker']

    def test_project_methods(self):
        """
        Test allowed methods.
        """
        response = self.get('/api/v1/project/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(['get'], json_data['allowed_detail_http_methods'])
        self.assertEqual(['get'], json_data['allowed_list_http_methods'])

    def test_projects_count(self):
        """
        Test that we have two projects in the database.
        """
        self.assertEqual(2, Project.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/project/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test project 1', json_data['objects'][0]['title'])

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/project/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test project 1', json_data['objects'][0]['title'])


class WorkerTestCase(ApiTestBase):
    """
    Tests for the worker interface.
    """
    fixtures = ['test_auth', 'test_project', 'test_worker']

    def test_worker_methods(self):
        """
        Test allowed methods.
        """
        response = self.get('/api/v1/worker/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(
            ['get', 'patch'], json_data['allowed_detail_http_methods'])
        self.assertEqual(['get'], json_data['allowed_list_http_methods'])

    def test_worker_count(self):
        """
        Test worker count in the DB.
        """
        self.assertEqual(2, Worker.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/worker/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test worker 1', json_data['objects'][0]['title'])

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/worker/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test worker 1', json_data['objects'][0]['title'])

    def test_patch_ping_response_dts(self):
        """
        Test PATCH the ping_response_dts field.
        """
        dts = timezone.now()
        response = self.patch(
            '/api/v1/worker/1/',
            {'ping_response_dts': dts.isoformat(' ')}
        )
        self.assertEqual(202, response.status_code)

        worker = Worker.objects.get(pk=1)
        self.assertEqual(dts, worker.ping_response_dts)


class JobTemplateTestCase(ApiTestBase):
    """
    Tests for the job-template interface.
    """
    fixtures = [
        'test_auth', 'test_project', 'test_worker', 'test_job_template']

    def test_job_template_methods(self):
        """
        Test allowed methods.
        """
        response = self.get('/api/v1/job_template/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(['get'], json_data['allowed_detail_http_methods'])
        self.assertEqual(['get'], json_data['allowed_list_http_methods'])

    def test_job_template_count(self):
        """
        Test job-template count in the db.
        """
        self.assertEqual(2, JobTemplate.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/job_template/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test template 1', json_data['objects'][0]['title'])

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/job_template/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test template 1', json_data['objects'][0]['title'])


class JobTestCase(ApiTestBase):
    """
    Tests for the job interface.
    """
    fixtures = [
        'test_auth',
        'test_project',
        'test_worker',
        'test_job_template',
        'test_job',
    ]

    def test_job_methods(self):
        """
        Test allowed methods on jobs.
        """
        response = self.get('/api/v1/job/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(
            ['get', 'put'], json_data['allowed_detail_http_methods'])
        self.assertEqual(['get'], json_data['allowed_list_http_methods'])

    def test_job_count(self):
        """
        Test the job count in the db.
        """
        self.assertEqual(2, Job.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/job/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test job 1', json_data['objects'][0]['title'])

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/job/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test job 1', json_data['objects'][0]['title'])


class RunTestCase(ApiTestBase):
    """
    Tests for the run interface.
    """
    fixtures = [
        'test_auth',
        'test_project',
        'test_worker',
        'test_job_template',
        'test_job',
    ]

    def test_run_methods(self):
        """
        Test allowed methods on runs.
        """
        response = self.get('/api/v1/run/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual([
            'get',
            'patch',
        ], json_data['allowed_detail_http_methods'])
        self.assertEqual(
            ['get', 'post'], json_data['allowed_list_http_methods'])

    def test_run_count(self):
        """
        Test the run count in the db.
        """
        self.assertEqual(2, Run.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/run/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual(1, json_data['objects'][0]['id'])

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/run/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual(1, json_data['objects'][0]['id'])

    def test_scheduled(self):
        """
        Test listing scheduled runs.
        """
        expected = [
            ('scheduled', 1),
            ('in_queue', 0),
            ('started', 0),
            ('completed', 0),
            ('completed_successful', 0),
            ('completed_with_error', 0),
        ]

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    def test_in_queue(self):
        """
        Test listing runs in queue.
        """
        expected = [
            ('scheduled', 0),
            ('in_queue', 1),
            ('started', 0),
            ('completed', 0),
            ('completed_successful', 0),
            ('completed_with_error', 0),
        ]

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.save()

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    def test_started(self):
        """
        Test listing runs that are started.
        """
        expected = [
            ('scheduled', 0),
            ('in_queue', 0),
            ('started', 1),
            ('completed', 0),
            ('completed_successful', 0),
            ('completed_with_error', 0),
        ]

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.start_dts = timezone.now()
        run.save()

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    def test_completed(self):
        """
        Test listing runs that are completed.
        """
        expected = [
            ('scheduled', 0),
            ('in_queue', 0),
            ('started', 0),
            ('completed', 1),
            ('completed_successful', 1),
            ('completed_with_error', 0),
        ]

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.start_dts = timezone.now()
        run.return_dts = timezone.now()
        run.return_success = True
        run.save()

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    def test_completed_with_error(self):
        """
        Test listing runs that are completed with error.
        """
        expected = [
            ('scheduled', 0),
            ('in_queue', 0),
            ('started', 0),
            ('completed', 1),
            ('completed_successful', 0),
            ('completed_with_error', 1),
        ]

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.start_dts = timezone.now()
        run.return_dts = timezone.now()
        run.return_success = False
        run.save()

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    def test_patch_run_1(self):
        """
        Test PATCH ``/api/v1/run/1/``.
        """
        enqueue_dts = timezone.now()

        response = self.patch(
            '/api/v1/run/1/',
            {
                'enqueue_dts': enqueue_dts.isoformat(' ')
            }
        )

        self.assertEqual(202, response.status_code)
        run = Run.objects.get(pk=1)
        self.assertEqual(enqueue_dts, run.enqueue_dts)

    def test_patch_with_reschedule(self):
        """
        Test PATCH ``/api/v1/run/1/`` causing a reschedule.
        """
        return_dts = timezone.now()
        Run.objects.update(enqueue_dts=timezone.now())
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': return_dts.isoformat(' '),
                'return_success': True,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertEqual(2, Run.objects.filter(job_id=1).count())
        self.assertEqual(
            return_dts, Run.objects.filter(job_id=1)[0].return_dts)

    def test_returned_with_error(self):
        """
        Test PATCH ``/api/v1/run/1/`` returned with error.

        This is expected to send e-mail notifications out.

        """
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': False,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertTrue(hasattr(mail, 'outbox'))
        self.assertEqual(1, len(mail.outbox))
        self.assertEqual(4, len(mail.outbox[0].to))
        self.assertEqual('Run error for: Test job 1', mail.outbox[0].subject)

    def test_returned_with_error_disable_enqueue(self):
        """
        Test PATCH ``/api/v1/run/1/`` error disables enqueue.

        This is expected to disable the enqueue.

        """
        job = Job.objects.get(pk=1)
        job.disable_enqueue_after_fails = 3
        job.save()

        for i in range(3):
            response = self.patch(
                '/api/v1/run/1/',
                {
                    'return_dts': timezone.now().isoformat(' '),
                    'return_success': False,
                }
            )

            self.assertEqual(202, response.status_code)
            job = Job.objects.get(pk=1)
            self.assertTrue(job.enqueue_is_enabled)
            self.assertEqual(i + 1, job.fail_times)

        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': False,
            }
        )

        job = Job.objects.get(pk=1)
        self.assertFalse(job.enqueue_is_enabled)

    def test_create_new_run(self):
        """
        Test create a new run by performing a ``POST``.
        """
        self.client.login(username='admin', password='admin')

        response = self.client.post(
            '/api/v1/run/',
            data=json.dumps({
                'job': '/api/v1/job/1/',
                'schedule_dts': '2013-01-01 00:00:00',
            }),
            content_type='application/json',
            ACCEPT='application/json',
        )

        self.assertEqual(201, response.status_code)
        run = Run.objects.filter(job_id=1)[1]
        self.assertEqual(
            timezone.make_aware(
                datetime(2013, 1, 1),
                timezone.get_default_timezone()
            ),
            run.schedule_dts
        )

    def test_create_new_run_no_permission(self):
        """
        Test create a new run by performing a ``POST`` with invalid group.
        """
        template = JobTemplate.objects.get(pk=1)
        template.auth_groups = [Group.objects.get(pk=2)]
        template.save()

        self.client.login(username='admin', password='admin')

        response = self.client.post(
            '/api/v1/run/',
            data=json.dumps({
                'job': '/api/v1/job/1/',
                'schedule_dts': '2013-01-01 00:00:00',
            }),
            content_type='application/json',
            ACCEPT='application/json',
        )

        self.assertEqual(400, response.status_code)


class ChainedRunTestCase(ApiTestBase):
    """
    Tests for the run interface with chaining.
    """
    fixtures = [
        'test_auth',
        'test_project',
        'test_worker',
        'test_job_template',
        'test_job',
        'test_child_job',
    ]

    def test_patch_with_reschedule(self):
        """
        Test PATCH ``/api/v1/run/1/`` for chained job.

        """
        Run.objects.update(enqueue_dts=timezone.now())
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': True,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertEqual(2, Job.objects.get(pk=1).run_set.count())
        self.assertEqual(1, Job.objects.get(pk=3).run_set.count())

    def test_patch_with_reschedule_no_children_reschedule(self):
        """
        Test PATCH ``/api/v1/run/1/`` for chained job but not reschedule
        children.

        """
        Run.objects.update(
            enqueue_dts=timezone.now(),
            schedule_children=False,
        )
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': True,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertEqual(2, Job.objects.get(pk=1).run_set.count())
        self.assertEqual(0, Job.objects.get(pk=3).run_set.count())


class KillRequestTestCase(ApiTestBase):
    """
    Tests for the kill request interface.
    """
    fixtures = [
        'test_auth',
        'test_project',
        'test_worker',
        'test_job_template',
        'test_job',
        'test_child_job',
    ]

    def test_post_new_kill_request(self):
        """
        Test POST ``/api/v1/kill_request/``.
        """
        response = self.post(
            '/api/v1/kill_request/',
            {
                'run': '/api/v1/run/1/',
            }
        )

        self.assertEqual(201, response.status_code)
        self.assertEqual(1, KillRequest.objects.filter(run_id=1).count())


class RunLogTestCase(ApiTestBase):
    """
    Tests for the run log interface.
    """
    fixtures = [
        'test_auth',
        'test_project',
        'test_worker',
        'test_job_template',
        'test_job',
        'test_child_job',
    ]

    def test_post_new_run_log(self):
        """
        Test POST ``/api/v1/run_log/``.
        """
        response = self.post(
            '/api/v1/run_log/',
            {
                'run': '/api/v1/run/1/',
                'content': 'foo bar',
            }
        )
        self.assertEqual(201, response.status_code)
        self.assertEqual(1, RunLog.objects.filter(run_id=1).count())



from django.test import TestCase

from mock import Mock, patch

from job_runner.apps.job_runner.auth import validate_hmac


class ModuleTestCase(TestCase):
    """
    Tests for :mod:`apps.job_runner.auth`.
    """
    @patch('job_runner.apps.job_runner.auth.Worker')
    def test_validate_hmac_valid_api_key(self, Worker):
        """
        Test :func:`.validate_hmac` with valid API key.
        """
        request = Mock()
        request.method = 'POST'
        request.META = {
            'HTTP_AUTHORIZATION': (
                'ApiKey api_user:9bfd7ed2a963182c4005c851f5f862cc97607f7b'),
        }
        request.get_full_path.return_value = 'full_request_path'
        request.raw_post_data = 'raw_post_data'

        api_user = Worker.objects.get.return_value
        api_user.secret = 'foobar'

        self.assertTrue(validate_hmac(request))

    @patch('job_runner.apps.job_runner.auth.Worker')
    def test_validate_hmac_invalid_api_key(self, Worker):
        """
        Test :func:`.validate_hmac` with invalid API key.
        """
        request = Mock()
        request.method = 'GET'
        request.META = {
            'HTTP_AUTHORIZATION': (
                'ApiKey api_user:9bfd7ed2a963182c4005c851f5f862cc97607f7b'),
        }
        request.get_full_path.return_value = 'full_request_path'
        request.raw_post_data = 'raw_post_data'

        api_user = Worker.objects.get.return_value
        api_user.secret = 'foobar'

        self.assertFalse(validate_hmac(request))



from django.test import TestCase
from django.utils import timezone

from job_runner.apps.job_runner.models import KillRequest, Run


class RunManagerTestCase(TestCase):
    """
    Tests for the job run manager.
    """
    fixtures = ['test_job']

    def test_scheduled(self):
        """
        Test :meth:`.Run.scheduled`.
        """
        self.assertEqual(
            1, Run.objects.scheduled().filter(job_id=1).count())

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.save()

        self.assertEqual(
            0, Run.objects.scheduled().filter(job_id=1).count())


class KillRequestManagerTestCase(TestCase):
    """
    Tests for the kill-request manager.
    """
    fixtures = ['test_job']

    def test_killable(self):
        """
        Test with one killable run.
        """
        run = Run.objects.get(pk=1)
        run.start_dts = timezone.now()
        run.pid = 1234
        run.save()

        KillRequest.objects.create(
            run=run,
            schedule_dts=timezone.now(),
        )

        self.assertEqual(1, KillRequest.objects.killable().count())

    def test_killable_run_not_yet_started(self):
        """
        Test when the run has not yet been started.
        """
        KillRequest.objects.create(
            run=Run.objects.get(pk=1),
            schedule_dts=timezone.now(),
        )

        self.assertEqual(0, KillRequest.objects.killable().count())

    def test_killable_run_already_returned(self):
        """
        Test whent he run has already been finished.
        """
        run = Run.objects.get(pk=1)
        run.start_dts = timezone.now()
        run.return_dts = timezone.now()
        run.pid = 1234
        run.save()

        KillRequest.objects.create(
            run=Run.objects.get(pk=1),
            schedule_dts=timezone.now(),
        )

        self.assertEqual(0, KillRequest.objects.killable().count())

    def test_killable_already_in_queue(self):
        """
        Test with one killable run which is already in queue.
        """
        run = Run.objects.get(pk=1)
        run.start_dts = timezone.now()
        run.pid = 1234
        run.save()

        KillRequest.objects.killable().create(
            run=run,
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
        )

        self.assertEqual(0, KillRequest.objects.killable().count())

    def test_killable_already_executed(self):
        """
        Test with one killable run which has been already executed.
        """
        run = Run.objects.get(pk=1)
        run.start_dts = timezone.now()
        run.pid = 1234
        run.save()

        KillRequest.objects.killable().create(
            run=run,
            schedule_dts=timezone.now(),
            execute_dts=timezone.now(),
        )

        self.assertEqual(0, KillRequest.objects.killable().count())



from django.conf.urls import patterns, include, url
from django.contrib import admin


admin.autodiscover()

urlpatterns = patterns(
    '',
    (r'^grappelli/', include('grappelli.urls')),
    url(r'^admin/', include(admin.site.urls)),

    (r'^api/', include('job_runner.apps.job_runner.api_urls')),

    url(r'^accounts/login/$', 'django.contrib.auth.views.login', name='login'),
    url(
        r'^accounts/logout/$', 'django.contrib.auth.views.logout',
        {'next_page': '/'},
        name='logout'
    ),

    (r'^', include('job_runner.apps.job_runner.urls', namespace='job_runner')),
)



import hashlib
import hmac
import re

from django.db.models import Q
from tastypie.authentication import Authentication
from tastypie.authorization import Authorization

from job_runner.apps.job_runner.models import Worker


def validate_hmac(request):
    """
    Validate the HMAC of an incoming request.

    :param request:
        The incoming request object.

    :return:
        A ``bool`` indicating if the request is valid.

    This will test the ``Authentication`` header for valid keys. The following
    format is expected::

        Authorization: ApiKey api_key:hmac_sha1

    ``public_key``
        Is the public key as stored in the ApiKey model.

    ``hmac_sha1``
        The HMAC-SHA1 generated over the uppercased request method + full
        request path (path + query string, if applicable) + request body,
        if applicable.

    """
    auth_header = request.META.get('HTTP_AUTHORIZATION', '')
    api_key_match = re.match(r'^ApiKey (.*?):(.*?)$', auth_header)

    if not api_key_match:
        return False

    try:
        worker = Worker.objects.get(api_key=api_key_match.group(1))
    except Worker.DoesNotExist:
        return False

    hmac_message = '{request_method}{full_path}{request_body}'.format(
        request_method=request.method,
        full_path=request.get_full_path(),
        request_body=request.raw_post_data,
    )

    expected_hmac = hmac.new(
        str(worker.secret), hmac_message, hashlib.sha1)

    if expected_hmac.hexdigest() == api_key_match.group(2):
        return True

    return False


class HmacAuthentication(Authentication):
    """
    Authenticate based on request HMAC.
    """
    def is_authenticated(self, request, **kwargs):
        return validate_hmac(request)


class ModelAuthorization(Authorization):
    """
    Model authorization.

    The purpose of this custom authorization class is to limit the number of
    results returned to the user based on session or API-key.

    In case of the session it will be limited to the groups the logged-in
    user is assiged to, in case of the API-key it will be limited to the
    data that refereces back to the API-key.

    :param api_key_path:
        The path relative from the used model to the API-key. For the
        ``JobTemplate`` this would be ``'worker__api_key'``.

    :param user_groups_path:
        The path relative from the used model to the groups the user belongs
        to. For the ``JobTemplate`` this would be
        ``'worker__project__groups'``.

    :param auth_user_groups_path:
        The path relative from the used model to the groups that are authorized
        to make modifications. Default is ``None``. For the ``Job`` model this
        would be ``'job_template__auth_groups'``.

    """
    def __init__(
            self, api_key_path, user_groups_path, auth_user_groups_path=None):
        self.api_key_path = api_key_path
        self.user_groups_path = user_groups_path
        self.auth_user_groups_path = auth_user_groups_path

    def apply_limits(self, request, object_list):
        if request and 'HTTP_AUTHORIZATION' in request.META:
            auth_header = request.META.get('HTTP_AUTHORIZATION')
            api_key_match = re.match(r'^ApiKey (.*?):(.*?)$', auth_header)
            return object_list.filter(
                **{self.api_key_path: api_key_match.group(1)}).distinct()

        elif (request and request.user.is_authenticated()
              and request.user.groups.count() > 0):
            groups_or = None

            if self.user_groups_path == "":
                path = 'name'
            else:
                path = '{0}__name'.format(self.user_groups_path)

            for group in request.user.groups.all():
                q_obj = Q(**{path: group.name})
                if not groups_or:
                    groups_or = q_obj
                else:
                    groups_or = groups_or | q_obj

            object_list = object_list.filter(groups_or).distinct()

            # apply extra filters when the request is not a GET
            if (request.method != 'GET' and
                    self.auth_user_groups_path is not None):
                groups_or = None

                if self.auth_user_groups_path == "":
                    path = 'name'
                else:
                    path = '{0}__name'.format(self.auth_user_groups_path)

                for group in request.user.groups.all():
                    q_obj = Q(**{path: group.name})
                    if not groups_or:
                        groups_or = q_obj
                    else:
                        groups_or = groups_or | q_obj

                object_list = object_list.filter(groups_or).distinct()

            return object_list

        return object_list.none()



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Project'
        db.create_table('job_runner_project', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('notification_addresses', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal('job_runner', ['Project'])

        # Adding M2M table for field groups on 'Project'
        db.create_table('job_runner_project_groups', (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('project', models.ForeignKey(orm['job_runner.project'], null=False)),
            ('group', models.ForeignKey(orm['auth.group'], null=False))
        ))
        db.create_unique('job_runner_project_groups', ['project_id', 'group_id'])

        # Adding model 'Worker'
        db.create_table('job_runner_worker', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('api_key', self.gf('django.db.models.fields.CharField')(unique=True, max_length=255, db_index=True)),
            ('secret', self.gf('django.db.models.fields.CharField')(max_length=255, db_index=True)),
            ('project', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['job_runner.Project'])),
            ('notification_addresses', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal('job_runner', ['Worker'])

        # Adding model 'JobTemplate'
        db.create_table('job_runner_jobtemplate', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('body', self.gf('django.db.models.fields.TextField')()),
            ('worker', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['job_runner.Worker'])),
            ('notification_addresses', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal('job_runner', ['JobTemplate'])

        # Adding M2M table for field auth_groups on 'JobTemplate'
        db.create_table('job_runner_jobtemplate_auth_groups', (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('jobtemplate', models.ForeignKey(orm['job_runner.jobtemplate'], null=False)),
            ('group', models.ForeignKey(orm['auth.group'], null=False))
        ))
        db.create_unique('job_runner_jobtemplate_auth_groups', ['jobtemplate_id', 'group_id'])

        # Adding model 'Job'
        db.create_table('job_runner_job', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('parent', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='children', null=True, to=orm['job_runner.Job'])),
            ('job_template', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['job_runner.JobTemplate'])),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('script_content_partial', self.gf('django.db.models.fields.TextField')()),
            ('script_content', self.gf('django.db.models.fields.TextField')()),
            ('reschedule_interval', self.gf('django.db.models.fields.PositiveIntegerField')(null=True, blank=True)),
            ('reschedule_interval_type', self.gf('django.db.models.fields.CharField')(max_length=6, blank=True)),
            ('reschedule_type', self.gf('django.db.models.fields.CharField')(max_length=18, blank=True)),
            ('notification_addresses', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal('job_runner', ['Job'])

        # Adding model 'RescheduleExclude'
        db.create_table('job_runner_rescheduleexclude', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('job', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['job_runner.Job'])),
            ('start_time', self.gf('django.db.models.fields.TimeField')()),
            ('end_time', self.gf('django.db.models.fields.TimeField')()),
        ))
        db.send_create_signal('job_runner', ['RescheduleExclude'])

        # Adding model 'Run'
        db.create_table('job_runner_run', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('job', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['job_runner.Job'])),
            ('schedule_dts', self.gf('django.db.models.fields.DateTimeField')(db_index=True)),
            ('enqueue_dts', self.gf('django.db.models.fields.DateTimeField')(null=True)),
            ('start_dts', self.gf('django.db.models.fields.DateTimeField')(null=True, db_index=True)),
            ('return_dts', self.gf('django.db.models.fields.DateTimeField')(null=True)),
            ('return_success', self.gf('django.db.models.fields.NullBooleanField')(default=None, null=True, blank=True)),
            ('return_log', self.gf('django.db.models.fields.TextField')(default=None, null=True)),
        ))
        db.send_create_signal('job_runner', ['Run'])


    def backwards(self, orm):
        # Deleting model 'Project'
        db.delete_table('job_runner_project')

        # Removing M2M table for field groups on 'Project'
        db.delete_table('job_runner_project_groups')

        # Deleting model 'Worker'
        db.delete_table('job_runner_worker')

        # Deleting model 'JobTemplate'
        db.delete_table('job_runner_jobtemplate')

        # Removing M2M table for field auth_groups on 'JobTemplate'
        db.delete_table('job_runner_jobtemplate_auth_groups')

        # Deleting model 'Job'
        db.delete_table('job_runner_job')

        # Deleting model 'RescheduleExclude'
        db.delete_table('job_runner_rescheduleexclude')

        # Deleting model 'Run'
        db.delete_table('job_runner_run')


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'job_runner.job': {
            'Meta': {'object_name': 'Job'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'job_template': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.JobTemplate']"}),
            'notification_addresses': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'children'", 'null': 'True', 'to': "orm['job_runner.Job']"}),
            'reschedule_interval': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'reschedule_interval_type': ('django.db.models.fields.CharField', [], {'max_length': '6', 'blank': 'True'}),
            'reschedule_type': ('django.db.models.fields.CharField', [], {'max_length': '18', 'blank': 'True'}),
            'script_content': ('django.db.models.fields.TextField', [], {}),
            'script_content_partial': ('django.db.models.fields.TextField', [], {}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'job_runner.jobtemplate': {
            'Meta': {'object_name': 'JobTemplate'},
            'auth_groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'body': ('django.db.models.fields.TextField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'notification_addresses': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'worker': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.Worker']"})
        },
        'job_runner.project': {
            'Meta': {'object_name': 'Project'},
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'notification_addresses': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'job_runner.rescheduleexclude': {
            'Meta': {'object_name': 'RescheduleExclude'},
            'end_time': ('django.db.models.fields.TimeField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'job': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.Job']"}),
            'start_time': ('django.db.models.fields.TimeField', [], {})
        },
        'job_runner.run': {
            'Meta': {'ordering': "('-id',)", 'object_name': 'Run'},
            'enqueue_dts': ('django.db.models.fields.DateTimeField', [], {'null': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'job': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.Job']"}),
            'return_dts': ('django.db.models.fields.DateTimeField', [], {'null': 'True'}),
            'return_log': ('django.db.models.fields.TextField', [], {'default': 'None', 'null': 'True'}),
            'return_success': ('django.db.models.fields.NullBooleanField', [], {'default': 'None', 'null': 'True', 'blank': 'True'}),
            'schedule_dts': ('django.db.models.fields.DateTimeField', [], {'db_index': 'True'}),
            'start_dts': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'db_index': 'True'})
        },
        'job_runner.worker': {
            'Meta': {'object_name': 'Worker'},
            'api_key': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255', 'db_index': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'notification_addresses': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'project': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.Project']"}),
            'secret': ('django.db.models.fields.CharField', [], {'max_length': '255', 'db_index': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        }
    }

    complete_apps = ['job_runner']


from datetime import datetime, time, timedelta

from django.core import mail
from django.test import TestCase
from django.utils import timezone

from job_runner.apps.job_runner.models import (
    Job,
    RescheduleExclude,
    Run,
)


class RunTestCase(TestCase):
    """
    Tests for the job run model.
    """
    fixtures = [
        'test_auth',
        'test_project',
        'test_worker',
        'test_job_template',
        'test_job'
    ]

    def test_reschedule_after_schedule_dts(self):
        """
        Test reschedule after schedule dts.
        """
        self.assertEqual(1, Run.objects.filter(job_id=1).count())
        Job.objects.get(pk=1).reschedule()
        self.assertEqual(1, Run.objects.filter(job_id=1).count())

        run = Run.objects.get(pk=1)
        run.schedule_dts = timezone.now()
        run.enqueue_dts = timezone.now()
        run.return_dts = timezone.now()
        run.save()

        Job.objects.get(pk=1).reschedule()
        self.assertEqual(2, Run.objects.filter(job_id=1).count())

        runs = Run.objects.filter(job_id=1).all()
        self.assertEqual(
            runs[0].schedule_dts + timedelta(days=1),
            runs[1].schedule_dts
        )

    def test_reschedule_after_schedule_dts_not_in_past(self):
        """
        Test reschedule after schedule dts is never in the past.
        """
        dts_now = timezone.now()

        run = Run.objects.get(pk=1)
        run.schedule_dts = dts_now - timedelta(days=31)
        run.enqueue_dts = dts_now - timedelta(days=31)
        run.return_dts = dts_now - timedelta(days=31)
        run.save()

        Job.objects.get(pk=1).reschedule()
        self.assertEqual(2, Run.objects.filter(job_id=1).count())

        run = Run.objects.get(pk=3)
        self.assertEqual(run.schedule_dts, dts_now + timedelta(days=1))

    def test_reschedule_after_complete_dts(self):
        """
        Test reschedule after complete dts.
        """
        job = Job.objects.get(pk=1)
        job.reschedule_type = 'AFTER_COMPLETE_DTS'
        job.save()

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.return_dts = timezone.now()
        run.save()

        job.reschedule()

        self.assertEqual(2, Run.objects.filter(job_id=1).count())

        runs = Run.objects.filter(job_id=1).all()
        self.assertEqual(
            runs[0].return_dts + timedelta(days=1),
            runs[1].schedule_dts
        )

    def test_reschedule_monthly_after_schedule_dts(self):
        """
        Test monthy reschedule after schedule dts.
        """
        in_and_expected = [
            (datetime(2032, 1, 1), datetime(2032, 2, 1)),
            (datetime(2032, 2, 1), datetime(2032, 3, 1)),
            (datetime(2032, 3, 1), datetime(2032, 4, 1)),
            (datetime(2032, 4, 1), datetime(2032, 5, 1)),
            (datetime(2032, 6, 1), datetime(2032, 7, 1)),
            (datetime(2032, 12, 1), datetime(2033, 1, 1)),
        ]

        for in_dts, expected_dts in in_and_expected:

            # if you do not use utc as a timezone here, you get funny results
            # because of day light saving time switching
            in_dts = timezone.make_aware(
                in_dts, timezone.utc)
            expected_dts = timezone.make_aware(
                expected_dts, timezone.utc)

            job = Job.objects.get(pk=1)
            job.reschedule_interval_type = 'MONTH'
            job.save()

            job.run_set.all().delete()

            Run.objects.create(
                job=job,
                schedule_dts=in_dts,
                enqueue_dts=in_dts,
                start_dts=in_dts,
                return_dts=in_dts,
            )

            job.reschedule()
            run = Run.objects.filter(job=1, enqueue_dts__isnull=True)[0]
            self.assertEqual(expected_dts, run.schedule_dts)

    def test_reschedule_with_exclude(self):
        """
        Test reschedule with exclude time.
        """
        job = Job.objects.get(pk=1)
        job.reschedule_type = 'AFTER_COMPLETE_DTS'
        job.reschedule_interval_type = 'HOUR'
        job.save()

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.return_dts = timezone.make_aware(
            datetime(2032, 1, 1, 11, 59), timezone.get_default_timezone())
        run.save()

        RescheduleExclude.objects.create(
            job=job,
            start_time=time(12, 00),
            end_time=time(13, 00),
        )

        job.reschedule()

        self.assertEqual(2, Run.objects.filter(job_id=1).count())

        runs = Run.objects.filter(job_id=1)
        self.assertEqual(
            timezone.make_aware(
                datetime(2032, 1, 1, 13, 59),
                timezone.get_default_timezone()
            ),
            runs[1].schedule_dts
        )

    def test_reschedule_with_invalid_exclude(self):
        """
        Test reschedule with exclude time which is invalid.
        """
        job = Job.objects.get(pk=1)
        job.reschedule_type = 'AFTER_COMPLETE_DTS'
        job.reschedule_interval_type = 'HOUR'
        job.save()

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.return_dts = timezone.make_aware(
            datetime(2012, 1, 1, 11, 59), timezone.get_default_timezone())
        run.save()

        RescheduleExclude.objects.create(
            job=job,
            start_time=time(0, 0),
            end_time=time(23, 59),
        )

        job.reschedule()

        self.assertEqual(1, Run.objects.filter(job_id=1).count())
        self.assertTrue(hasattr(mail, 'outbox'))
        self.assertEqual(1, len(mail.outbox))
        self.assertEqual(4, len(mail.outbox[0].to))
        self.assertEqual(
            'Reschedule error for: Test job 1', mail.outbox[0].subject)

    def test_reschedule_with_run_scheduled(self):
        """
        Test reschedule when there is already a run scheduled.
        """
        job = Job.objects.get(pk=1)
        self.assertEqual(1, Run.objects.filter(job=job).count())

        run = Run.objects.get(pk=1)
        run.schedule_dts = timezone.now()
        run.enqueue_dts = timezone.now()
        run.return_dts = timezone.now()
        run.save()

        Run.objects.create(
            job=job,
            schedule_dts=timezone.now()
        )

        self.assertEqual(2, Run.objects.filter(job=job).count())
        job.reschedule()
        self.assertEqual(2, Run.objects.filter(job=job).count())

    def test_reschedule_with_started_run(self):
        """
        Test reschedule when there is already an other started run active.
        """
        job = Job.objects.get(pk=1)
        self.assertEqual(1, Run.objects.filter(job=job).count())

        run = Run.objects.get(pk=1)
        run.schedule_dts = timezone.now()
        run.enqueue_dts = timezone.now()
        run.start_dts = timezone.now()
        run.return_dts = timezone.now()
        run.save()

        Run.objects.create(
            job=job,
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
            start_dts=timezone.now(),
        )

        self.assertEqual(2, Run.objects.filter(job=job).count())
        job.reschedule()
        self.assertEqual(2, Run.objects.filter(job=job).count())

    def test_get_notification_addresses(self):
        """
        Test ``get_notification_addresses`` methods on models.
        """
        self.assertEqual(
            sorted([
                'project@example.com',
                'worker@example.com',
                'template@example.com',
                'job@example.com',
            ]),
            sorted(Job.objects.get(pk=1).get_notification_addresses())
        )

    def test_schedule_now(self):
        """
        Test direct schedule.
        """
        Run.objects.all().delete()

        job = Job.objects.get(pk=1)

        self.assertEqual(0, job.run_set.count())
        job.schedule_now()
        self.assertEqual(1, job.run_set.count())



from django.db import models
from django.db.models import Q
from django.utils import timezone


class RunManager(models.Manager):
    """
    Custom manager for the Run model.
    """
    def scheduled(self):
        """
        Return a QS filtered on run's are in the scheduled state.
        """
        qs = self.get_query_set()
        return qs.filter(enqueue_dts__isnull=True)

    def enqueueable(self):
        """
        Return a QS filtered on runs's that are ready to enqueue.
        """
        job_qs = self.model.job.get_query_set()

        active_jobs = job_qs.filter(
            run__enqueue_dts__isnull=False,
            run__return_dts__isnull=True,
        )
        scheduled = self.scheduled()

        return scheduled.filter(
            # make sure it should be running now
            schedule_dts__lte=timezone.now(),
        ).exclude(
            # exclude auto scheduled jobs when enqueue is disabled
            Q(
                job__enqueue_is_enabled=False,
                is_manual=False
            ) |
            Q(
                job__job_template__enqueue_is_enabled=False,
                is_manual=False
            ) |
            Q(
                job__job_template__worker__enqueue_is_enabled=False,
                is_manual=False
            ) |
            Q(
                job__job_template__worker__project__enqueue_is_enabled=False,
                is_manual=False
            ) |

            # exclude jobs that are still active
            Q(job__in=active_jobs)
        )


class KillRequestManager(models.Manager):
    """
    Custom manager for the KillRequest model.
    """
    def killable(self):
        """
        Return a QS filtered on requests that are killable.
        """
        qs = self.get_query_set()

        return qs.filter(
            # this should be always the case
            schedule_dts__lte=timezone.now(),

            # make sure it hasn't been executed already
            enqueue_dts__isnull=True,
            execute_dts__isnull=True,

            # make sure a pid is assigned to the run
            run__pid__isnull=False,

            # make sure the run is active
            run__start_dts__isnull=False,

            # make sure the run hasn't returned already
            run__return_dts__isnull=True,
        )



from django.conf.urls.defaults import include, patterns
from tastypie.api import Api

from job_runner.apps.job_runner.api import (
    GroupResource,
    JobResource,
    JobTemplateResource,
    KillRequestResource,
    ProjectResource,
    RunLogResource,
    RunResource,
    WorkerResource,
)


v1_api = Api(api_name='v1')
v1_api.register(GroupResource()),
v1_api.register(JobResource())
v1_api.register(JobTemplateResource())
v1_api.register(KillRequestResource())
v1_api.register(ProjectResource())
v1_api.register(RunLogResource())
v1_api.register(RunResource())
v1_api.register(WorkerResource())

urlpatterns = patterns(
    '',
    (r'', include(v1_api.urls)),
)



from django.contrib.auth.models import Group
from tastypie import fields
from tastypie.authentication import MultiAuthentication, SessionAuthentication
from tastypie.constants import ALL, ALL_WITH_RELATIONS
from tastypie.resources import ModelResource

from job_runner.apps.job_runner.auth import (
    HmacAuthentication, ModelAuthorization)
from job_runner.apps.job_runner.models import (
    Job,
    JobTemplate,
    KillRequest,
    Project,
    Run,
    RunLog,
    Worker,
)


class NoRelatedSaveMixin(object):
    def save_related(self, *args, **kwargs):
        """
        Override to NOT save related models.

        In this case there is no need for it (it will update the related run
        every time a kill-requests is updated). Case where it goes wrong:

        * Kill requests has been executed -> PATCH kill_request
        * Run returned                    -> PATCH run

        Since the PATCH kill_request will do a select on the run (and getting
        the old data), around the same time the PATCH run will update the run
        and around the same time the PATCH kill_request will update the run
        as well, we will end-up with a not-updated run.

        An other option to work around this is to use the SERIALIZE transaction
        level. However, this seems not to be possible with all Django DB
        backends.

        See: https://github.com/toastdriven/django-tastypie/issues/742

        """
        pass


class GroupResource(ModelResource):
    """
    RESTful resource for Django groups.
    """
    class Meta:
        queryset = Group.objects.all()
        resource_name = 'group'
        allowed_methods = ['get']
        fields = ['name']

        authentication = MultiAuthentication(
            SessionAuthentication(), HmacAuthentication())

        authorization = ModelAuthorization(
            api_key_path='jobtemplate__worker__api_key',
            user_groups_path='',
        )


class ProjectResource(ModelResource):
    """
    RESTful resource for projects.
    """
    class Meta:
        queryset = Project.objects.all()
        resource_name = 'project'
        allowed_methods = ['get']
        fields = ['title', 'id', 'description', 'enqueue_is_enabled']
        filtering = {
            'id': ('exact',),
        }

        authentication = MultiAuthentication(
            SessionAuthentication(), HmacAuthentication())

        authorization = ModelAuthorization(
            api_key_path='worker__api_key',
            user_groups_path='groups',
        )


class WorkerResource(ModelResource):
    """
    RESTful resource for workers.
    """
    project = fields.ToOneField(
        'job_runner.apps.job_runner.api.ProjectResource', 'project')

    class Meta:
        queryset = Worker.objects.all()
        resource_name = 'worker'
        list_allowed_methods = ['get']
        detail_allowed_methods = ['get', 'patch']
        fields = [
            'id',
            'title',
            'api_key',
            'description',
            'enqueue_is_enabled',
            'ping_response_dts'
        ]
        filtering = {
            'project': ALL_WITH_RELATIONS,
            'api_key': ALL,
        }

        authentication = MultiAuthentication(
            SessionAuthentication(), HmacAuthentication())

        authorization = ModelAuthorization(
            api_key_path='api_key',
            user_groups_path='project__groups',
        )


class JobTemplateResource(ModelResource):
    """
    RESTful resource for job-templates.
    """
    worker = fields.ToOneField(
        'job_runner.apps.job_runner.api.WorkerResource', 'worker')
    auth_groups = fields.ToManyField(
        'job_runner.apps.job_runner.api.GroupResource',
        'auth_groups',
        null=True
    )

    def build_filters(self, filters=None):
        if filters is None:
            filters = {}
        orm_filters = super(JobTemplateResource, self).build_filters(filters)

        if 'project_id' in filters:
            orm_filters.update({
                'worker__project__id': filters['project_id']
            })

        return orm_filters

    class Meta:
        queryset = JobTemplate.objects.all()
        resource_name = 'job_template'
        allowed_methods = ['get']
        fields = ['id', 'title', 'description', 'enqueue_is_enabled']
        filtering = {
            'worker': ALL_WITH_RELATIONS,
        }

        authentication = MultiAuthentication(
            SessionAuthentication(), HmacAuthentication())

        authorization = ModelAuthorization(
            api_key_path='worker__api_key',
            user_groups_path='worker__project__groups',
        )


class JobResource(NoRelatedSaveMixin, ModelResource):
    """
    RESTful resource for jobs.
    """
    job_template = fields.ToOneField(
        'job_runner.apps.job_runner.api.JobTemplateResource', 'job_template')
    parent = fields.ToOneField('self', 'parent', null=True)
    children = fields.ToManyField('self', 'children', null=True)

    def build_filters(self, filters=None):
        if filters is None:
            filters = {}
        orm_filters = super(JobResource, self).build_filters(filters)

        if 'project_id' in filters:
            orm_filters.update({
                'job_template__worker__project__id': filters['project_id']
            })

        return orm_filters

    class Meta:
        queryset = Job.objects.all()
        resource_name = 'job'
        detail_allowed_methods = ['get', 'put']
        list_allowed_methods = ['get']
        fields = [
            'id',
            'title',
            'description',
            'script_content',
            'enqueue_is_enabled',
            'reschedule_interval',
            'reschedule_interval_type',
            'reschedule_type',
        ]
        filtering = {
            'job_template': ALL_WITH_RELATIONS,
        }

        authentication = MultiAuthentication(
            SessionAuthentication(), HmacAuthentication())

        authorization = ModelAuthorization(
            api_key_path='job_template__worker__api_key',
            user_groups_path='job_template__worker__project__groups',
            auth_user_groups_path='job_template__auth_groups',
        )


class RunResource(NoRelatedSaveMixin, ModelResource):
    """
    RESTful resource for job runs.
    """
    job = fields.ToOneField(
        'job_runner.apps.job_runner.api.JobResource', 'job')

    run_log = fields.ToOneField(
        'job_runner.apps.job_runner.api.RunLogResource', 'run_log',
        null=True,
        blank=True
    )

    class Meta:
        queryset = Run.objects.filter()
        resource_name = 'run'
        detail_allowed_methods = ['get', 'patch']
        list_allowed_methods = ['get', 'post']
        filtering = {
            'schedule_dts': ALL,
            'job': ALL_WITH_RELATIONS,
        }

        authentication = MultiAuthentication(
            SessionAuthentication(), HmacAuthentication())

        authorization = ModelAuthorization(
            api_key_path='job__job_template__worker__api_key',
            user_groups_path='job__job_template__worker__project__groups',
            auth_user_groups_path='job__job_template__auth_groups',
        )

    def build_filters(self, filters=None):
        if filters is None:
            filters = {}
        orm_filters = super(RunResource, self).build_filters(filters)

        state_filters = {
            'scheduled': {
                'enqueue_dts__isnull': True,
            },
            'in_queue': {
                'enqueue_dts__isnull': False,
                'start_dts__isnull': True,
            },
            'started': {
                'enqueue_dts__isnull': False,
                'start_dts__isnull': False,
                'return_dts__isnull': True,
            },
            'completed': {
                'enqueue_dts__isnull': False,
                'start_dts__isnull': False,
                'return_dts__isnull': False,
            },
            'completed_successful': {
                'enqueue_dts__isnull': False,
                'start_dts__isnull': False,
                'return_dts__isnull': False,
                'return_success': True,
            },
            'completed_with_error': {
                'enqueue_dts__isnull': False,
                'start_dts__isnull': False,
                'return_dts__isnull': False,
                'return_success': False,
            }
        }

        if 'state' in filters and filters['state'] in state_filters:
            orm_filters.update(state_filters[filters['state']])

        if 'project_id' in filters:
            orm_filters.update({
                'job__job_template__worker__project__id': filters['project_id']
            })

        return orm_filters

    def obj_update(self, bundle, request=None, *args, **kwargs):
        """
        Override of the default obj_update method.

        This will call the ``reschedule`` method after a successful object
        update, which will re-schedule the job if needed (incl. children).

        If the object update represents the returning of a run with error,
        it will call also ``send_error_notification`` method.

        """
        deserialized = self.deserialize(
            request,
            request.raw_post_data,
            format=request.META.get('CONTENT_TYPE', 'application/json')
        )

        result = super(RunResource, self).obj_update(
            bundle, request, *args, **kwargs)

        job = bundle.obj.job

        if (deserialized.get('return_dts', None) and
                deserialized.get('return_success', None) is False):
            # the job failed
            bundle.obj.send_error_notification()
            job.fail_times = job.fail_times + 1

            # disable job when it failed more than x times
            if (job.disable_enqueue_after_fails and
                    job.fail_times > job.disable_enqueue_after_fails):
                job.enqueue_is_enabled = False

            job.save()

        job.reschedule()

        if (deserialized.get('return_dts', None) and
                deserialized.get('return_success', None) is True):
            # reset the fail count
            job.fail_times = 0
            job.save()

        if (deserialized.get('return_dts', None) and
                deserialized.get('return_success', None) is True and
                bundle.obj.schedule_children):
            # the job completed successfully and has children to
            # schedule now
            for child in bundle.obj.job.children.all():
                child.schedule_now()

        return result


class KillRequestResource(NoRelatedSaveMixin, ModelResource):
    """
    RESTful resource for kill requests.
    """
    run = fields.ToOneField(
        'job_runner.apps.job_runner.api.RunResource', 'run')

    class Meta:
        queryset = KillRequest.objects.all()
        resource_name = 'kill_request'
        list_allowed_methods = ['get', 'post']
        detail_allowed_methods = ['get', 'patch']

        authentication = MultiAuthentication(
            SessionAuthentication(), HmacAuthentication())

        authorization = ModelAuthorization(
            api_key_path='run__job__job_template__worker__api_key',
            user_groups_path='run__job__job_template__worker__project__groups',
            auth_user_groups_path='run__job__job_template__auth_groups',
        )


class RunLogResource(NoRelatedSaveMixin, ModelResource):
    """
    RESTful resource for run log-output.
    """
    run = fields.ToOneField(
        'job_runner.apps.job_runner.api.RunResource', 'run')

    class Meta:
        queryset = RunLog.objects.all()
        resource_name = 'run_log'
        list_allowed_methods = ['get', 'post']
        detail_allowed_methods = ['get', 'patch']

        authentication = MultiAuthentication(
            SessionAuthentication(), HmacAuthentication())

        authorization = ModelAuthorization(
            api_key_path='run__job__job_template__worker__api_key',
            user_groups_path='run__job__job_template__worker__project__groups',
            auth_user_groups_path='run__job__job_template__auth_groups',
        )



from datetime import timedelta

from django.conf import settings
from django.contrib import admin
from django.db.models import Q
from django.utils import timezone
from django.utils.dateformat import DateFormat

from job_runner.apps.job_runner.models import (
    Job,
    JobTemplate,
    Project,
    RescheduleExclude,
    Run,
    Worker,
)


class PermissionAdminMixin(object):
    """
    Mixin class to limit the number of visible items.
    """

    groups_path = None
    """
    A ``str`` containing the path to the groups.

    Example::

        'job_template__auth_groups'

    """

    fk_groups_path = {}
    """
    A ``dict`` containing the FK field name and as a value the corresponding
    model and path to the groups.

    Example::

        {
            'job_template': {
                'path': 'auth_groups',
                'model': JobTemplate,
            }
        }

    """

    def _filter_groups(self, request, qs, groups_path):
        if request.user.is_superuser:
            return qs

        if not request.user.groups.count():
            return qs.none()

        groups_or = None
        for group in request.user.groups.all():
            q_obj = Q(**{groups_path: group})
            if not groups_or:
                groups_or = q_obj
            else:
                groups_or = groups_or | q_obj

        return qs.filter(groups_or).distinct()

    def queryset(self, request):
        """
        Get results based on groups the user is in.

        If the user is a super-user, the user has access to everything. Else,
        the results are limited based on matching groups as set in the model.

        """
        qs = super(PermissionAdminMixin, self).queryset(request)
        return self._filter_groups(request, qs, self.groups_path)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name in self.fK_groups_path:
            qs = self.fK_groups_path[db_field.name]['model'].objects.all()
            group_path = self.fK_groups_path[db_field.name]['path']
            kwargs['queryset'] = self._filter_groups(request, qs, group_path)

        return super(PermissionAdminMixin, self).formfield_for_foreignkey(
            db_field, request, **kwargs)


class RunInlineAdmin(admin.TabularInline):
    """
    Inline admin for job runs.
    """
    model = Run
    fields = ['schedule_dts']
    max_num = 1

    def queryset(self, request):
        return self.model.objects.scheduled()


class RescheduleExcludeInlineAdmin(admin.TabularInline):
    """
    Inline admin for reschedule excludes.
    """
    model = RescheduleExclude


class ProjectAdmin(admin.ModelAdmin):
    """
    Admin interface for projects.
    """
    list_display = ('title', 'enqueue_is_enabled')
    list_filter = ('enqueue_is_enabled',)
    list_editable = ('enqueue_is_enabled',)


class WorkerAdmin(admin.ModelAdmin):
    """
    Admin interface for workers.
    """
    list_display = (
        'project',
        'title',
        'api_key',
        'enqueue_is_enabled',
        'ping_response'
    )
    list_filter = ('project', 'enqueue_is_enabled')
    list_display_links = ('title',)
    list_editable = ('enqueue_is_enabled',)

    def ping_response(self, obj):
        interval = settings.JOB_RUNNER_WORKER_PING_INTERVAL
        margin = settings.JOB_RUNNER_WORKER_PING_MARGIN

        if not obj.ping_response_dts:
            return '<span style="color: red;">(none)</span>'

        date_local = obj.ping_response_dts.astimezone(
            timezone.get_current_timezone())
        date_formatted = DateFormat(date_local)
        date_formatted = date_formatted.format(settings.DATETIME_FORMAT)

        if (obj.ping_response_dts <
                timezone.now() - timedelta(seconds=2 * interval + margin)):
            return (
                '<img src="{0}admin/img/icon_error.gif" /> '
                '<span style="color: red;">{1}</span>'.format(
                    settings.STATIC_URL,
                    date_formatted
                )
            )

        if (obj.ping_response_dts <
                timezone.now() - timedelta(seconds=interval + margin)):
            return (
                '<img src="{0}admin/img/icon_alert.gif" /> '
                '<span style="color: orange;">{1}</span>'.format(
                    settings.STATIC_URL,
                    date_formatted
                )
            )

        return (
            '<img src="{0}admin/img/icon_success.gif" /> '
            '<span style="color: green;">{1}</span>'.format(
                settings.STATIC_URL,
                date_formatted
            )
        )

    ping_response.allow_tags = True
    ping_response.short_description = 'Last ping response'


class JobTemplateAdmin(admin.ModelAdmin):
    """
    Admin interface for job-templates.
    """
    list_display = ('worker', 'title', 'enqueue_is_enabled')
    list_filter = ('worker', 'worker__project', 'enqueue_is_enabled')
    list_display_links = ('title',)
    list_editable = ('enqueue_is_enabled',)


class JobAdmin(PermissionAdminMixin, admin.ModelAdmin):
    """
    Admin interface for jobs.
    """
    list_display = (
        'job_template',
        'title',
        'enqueue_is_enabled',
        'reschedule_type',
        'parent'
    )
    list_filter = (
        'enqueue_is_enabled',
        'reschedule_interval_type',
        'reschedule_type',
        'job_template',
        'job_template__worker',
        'job_template__worker__project'
    )
    list_display_links = ('title',)
    list_editable = ('enqueue_is_enabled',)
    search_fields = ('title',)
    inlines = [
        RunInlineAdmin,
        RescheduleExcludeInlineAdmin,
    ]
    groups_path = 'job_template__auth_groups'
    fK_groups_path = {
        'job_template': {
            'model': JobTemplate,
            'path': 'auth_groups',
        },
        'parent': {
            'model': Job,
            'path': 'job_template__auth_groups',
        }
    }

    fieldsets = (
        (None, {
            'fields': ('title', 'job_template', 'parent', 'description',)
        }),
        ('Script', {
            'fields': ('script_content_partial',)
        }),
        ('Notifications', {
            'fields': ('notification_addresses',)
        }),
        ('Scheduling', {
            'fields': (
                'enqueue_is_enabled',
                'disable_enqueue_after_fails',
                'reschedule_interval',
                'reschedule_interval_type',
                'reschedule_type',
            ),
        }),
    )


admin.site.register(Job, JobAdmin)
admin.site.register(JobTemplate, JobTemplateAdmin)
admin.site.register(Project, ProjectAdmin)
admin.site.register(Worker, WorkerAdmin)



from django.conf import settings
from django.views.generic import TemplateView


class DashboardView(TemplateView):
    template_name = 'job_runner/index.html'

    def get_context_data(self, **kwargs):
        context_data = super(DashboardView, self).get_context_data(**kwargs)
        context_data['ws_server'] = settings.JOB_RUNNER_WS_SERVER
        return context_data



from django.conf.urls.defaults import patterns, url
from django.contrib.auth.decorators import login_required

from job_runner.apps.job_runner.views import DashboardView


view = login_required(DashboardView.as_view())


urlpatterns = patterns(
    '',
    url(
        r'^project/(?P<project_id>\d+)/$',
        view,
        name='project'
    ),
    url(
        r'^project/(?P<project_id>\d+)/runs/$',
        view,
        name='runs'
    ),
    url(
        r'^project/(?P<project_id>\d+)/runs/(?P<run_id>\d+)/$',
        view,
        name='runs_run'
    ),
    url(
        r'^project/(?P<project_id>\d+)/jobs/$',
        view,
        name='jobs'
    ),
    url(
        r'^project/(?P<project_id>\d+)/jobs/(?P<job_id>\d+)/$',
        view,
        name='job'
    ),
    url(
        (
            r'^project/(?P<project_id>\d+)/jobs/'
            r'(?P<job_id>\d+)/runs/(?P<run_id>\d+)/$'
        ),
        view,
        name='job_run'
    ),
    url(
        r'^$',
        view,
        name='index'
    ),
)



import os

from job_runner.settings.base import *


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(PROJECT_ROOT, '..', 'test_database.sqlite'),
    }
}

JOB_RUNNER_WS_SERVER = 'ws://localhost:5000/'

TEST_RUNNER = 'django_nose.NoseTestSuiteRunner'


# Make sure that test with naive datetime objects will fail
import warnings
warnings.filterwarnings(
    'error',
    r"DateTimeField received a naive datetime",
    RuntimeWarning,
    r'django\.db\.models\.fields'
)



import json
import logging
import time
from datetime import datetime, timedelta

import zmq
from django.conf import settings
from django.core.management.base import NoArgsCommand

from job_runner.apps.job_runner.models import KillRequest, Run, Worker


logger = logging.getLogger(__name__)


class Command(NoArgsCommand):
    help = 'Broadcast runs and kill-requests to workers'

    def handle_noargs(self, **options):
        logger.info('Starting queue broadcaster')
        context = zmq.Context(1)
        publisher = context.socket(zmq.PUB)
        publisher.bind(
            'tcp://*:{0}'.format(settings.JOB_RUNNER_BROADCASTER_PORT))

        # give the subscribers some time to (re-)connect.
        time.sleep(2)

        ping_delta = timedelta(
            seconds=settings.JOB_RUNNER_WORKER_PING_INTERVAL)
        next_ping_request = datetime.utcnow()

        while True:
            if next_ping_request <= datetime.utcnow():
                self._broadcast_worker_ping(publisher)
                next_ping_request = datetime.utcnow() + ping_delta
            self._broadcast_runs(publisher)
            self._broadcast_kill_requests(publisher)
            time.sleep(5)

        publisher.close()
        context.term()

    def _broadcast_runs(self, publisher):
        """
        Broadcast runs that are scheduled to run now.

        When the job has ``job__enqueue_is_enabled`` set to ``False``, its
        runs are not broadcasted, unless they are scheduled manually
        (``is_manual`` set to ``True``).

        :param publisher:
            A ``zmq`` publisher.

        """
        enqueueable_runs = Run.objects.enqueueable().select_related()

        broadcasted_job_ids = []

        for run in enqueueable_runs:
            if run.job.pk not in broadcasted_job_ids:
                worker = run.job.job_template.worker
                message = [
                    'master.broadcast.{0}'.format(worker.api_key),
                    json.dumps({'run_id': run.id, 'action': 'enqueue'})
                ]
                logger.debug('Sending: {0}'.format(message))
                publisher.send_multipart(message)

                # in raw sql, this can be avoided, I couldn't find a way how
                # this can be done with the django orm
                broadcasted_job_ids.append(run.job.pk)

    def _broadcast_kill_requests(self, publisher):
        """
        Broadcast kill-requests.

        :param publisher:
            A ``zmq`` publisher.

        """
        kill_requests = KillRequest.objects.killable().select_related()

        for kill_request in kill_requests:
            run = kill_request.run
            worker = run.job.job_template.worker
            message = [
                'master.broadcast.{0}'.format(worker.api_key),
                json.dumps({
                    'kill_request_id': kill_request.id,
                    'action': 'kill',
                })
            ]
            logger.debug('Sending: {0}'.format(message))
            publisher.send_multipart(message)

    def _broadcast_worker_ping(self, publisher):
        """
        Broadcast ping-request to all the workers.

        :param publisher:
            A ``zmq`` publisher.

        """
        workers = Worker.objects.all()

        for worker in workers:
            message = [
                'master.broadcast.{0}'.format(worker.api_key),
                json.dumps({
                    'action': 'ping',
                })
            ]
            logger.debug('Sending: {0}'.format(message))
            publisher.send_multipart(message)



import calendar
import copy
from datetime import timedelta

from django.conf import settings
from django.core.mail import send_mail
from django.contrib.auth.models import Group
from django.db import models
from django.template import Context, Template
from django.template.loader import get_template
from django.utils import timezone

from job_runner.apps.job_runner.managers import KillRequestManager, RunManager


RESCHEDULE_INTERVAL_TYPE_CHOICES = (
    ('MINUTE', 'Every x minutes'),
    ('HOUR', 'Every x hours'),
    ('DAY', 'Every x days'),
    ('MONTH', 'Every x months'),
)

RESCHEDULE_TYPE_CHOICES = (
    ('AFTER_SCHEDULE_DTS', 'Increment schedule dts by interval'),
    ('AFTER_COMPLETE_DTS', 'Increment complete dts by interval'),
)


class RescheduleException(Exception):
    """
    Exception used when a rescheduling failed.
    """


class Project(models.Model):
    """
    Projects
    """
    title = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    groups = models.ManyToManyField(
        Group,
        help_text=(
            'These are the groups that can see the project in the dashboard. '
            'The admin permissions are set on template level. '

        )
    )
    notification_addresses = models.TextField(
        help_text='Separate e-mail addresses by a newline',
        blank=True,
    )
    enqueue_is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=(
            'If unchecked, nothing for this project will be added to '
            'the worker queue. This will not affect already running jobs.'
        )
    )

    def __unicode__(self):
        return self.title

    def get_notification_addresses(self):
        """
        Return a ``list`` notification addresses.
        """
        addresses = self.notification_addresses.strip().split('\n')
        return [x.strip() for x in addresses if x.strip() != '']

    class Meta:
        ordering = ('title', )


class Worker(models.Model):
    """
    Workers
    """
    title = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    api_key = models.CharField(max_length=255, db_index=True, unique=True)
    secret = models.CharField(max_length=255, db_index=True)
    project = models.ForeignKey(Project)
    notification_addresses = models.TextField(
        help_text='Separate e-mail addresses by a newline',
        blank=True,
    )
    enqueue_is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=(
            'If unchecked, nothing for this worker will be added to '
            'the worker queue. This will not affect already running jobs.'
        )
    )
    ping_response_dts = models.DateTimeField(
        blank=True, null=True, editable=False)

    def __unicode__(self):
        return u'{0} > {1}'.format(self.project, self.title)

    def get_notification_addresses(self):
        """
        Return a ``list`` of notification addresses.
        """
        addresses = self.notification_addresses.strip().split('\n')
        addresses = [x.strip() for x in addresses if x.strip() != '']
        addresses.extend(self.project.get_notification_addresses())
        return addresses

    class Meta:
        ordering = ('project__title', 'title', )


class JobTemplate(models.Model):
    """
    Job templates
    """
    title = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    body = models.TextField(help_text=(
        'Use {{ content|safe }} at the place where you want to render the '
        'script content of the job'
    ))
    worker = models.ForeignKey(Worker)
    auth_groups = models.ManyToManyField(
        Group,
        blank=True,
        help_text=(
            'These are the groups that are authorized to see this template '
            'and its jobs in the admin and are able to re-schedule the jobs '
            'using this template in the dashboard.'
        )
    )
    notification_addresses = models.TextField(
        help_text='Separate e-mail addresses by a newline',
        blank=True,
    )
    enqueue_is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=(
            'If unchecked, nothing for this template will be added to '
            'the worker queue. This will not affect already running jobs.'
        )
    )

    def __unicode__(self):
        return u'{0} > {1}'.format(self.worker, self.title)

    def save(self, *args, **kwargs):
        """
        Override default save to re-save jobs.

        This will make sure that when the template changed, the cached version
        will be updated on the job level.

        """
        super(JobTemplate, self).save(*args, **kwargs)
        for job in self.job_set.all():
            job.save()

    def get_notification_addresses(self):
        """
        Return a ``list`` of notification addresses.
        """
        addresses = self.notification_addresses.strip().split('\n')
        addresses = [x.strip() for x in addresses if x.strip() != '']
        addresses.extend(self.worker.get_notification_addresses())
        return addresses

    class Meta:
        ordering = ('worker__project__title', 'worker__title', 'title', )


class Job(models.Model):
    """
    Contains the job data.
    """
    parent = models.ForeignKey(
        'self', blank=True, null=True, related_name='children')
    job_template = models.ForeignKey(JobTemplate)
    title = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    script_content_partial = models.TextField('script content')
    script_content = models.TextField(editable=False)
    enqueue_is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=(
            'If unchecked, the job will not be added to the worker queue. '
            'This will not affect already running jobs.'
        )
    )
    reschedule_interval = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text=('Leave blank if you do not want to re-schedule this job.')
    )
    reschedule_interval_type = models.CharField(
        max_length=6,
        blank=True,
        choices=RESCHEDULE_INTERVAL_TYPE_CHOICES,
        db_index=True,
        help_text=(
            'Note: with montly re-scheduling, the date will be incremented '
            'by the number of days in the month.'
        )
    )
    reschedule_type = models.CharField(
        max_length=18,
        blank=True,
        choices=RESCHEDULE_TYPE_CHOICES,
        db_index=True,
    )
    notification_addresses = models.TextField(
        help_text='Separate addresses by a newline',
        blank=True,
    )
    fail_times = models.PositiveIntegerField(
        editable=False,
        default=0,
        help_text='The number of times this job failed in a row.',
    )
    disable_enqueue_after_fails = models.PositiveIntegerField(
        blank=True,
        null=True,
        help_text=(
            'The number of times after which the enqueue-ing of a job should '
            'be disabled when it failed (blank = never disable enqeueue).'
        )
    )

    class Meta:
        ordering = (
            'job_template__worker__project__title',
            'job_template__worker__title',
            'job_template__title', 'title',
        )
        unique_together = (('title', 'job_template'),)

    def __unicode__(self):
        return u'{0} > {1}'.format(self.job_template, self.title)

    def reschedule(self):
        """
        Reschedule job.

        This will check if the job is setup for reschedule. Then it will
        try to get the next reschedule date. When getting this date fails
        (eg: when the rescheduling always falls within the reschedule exclude),
        it will send out an e-mail to the Job-Runner admins and the e-mail
        addresses that are setup for this job, script and server.

        """
        # there is already an other run which is not finished yet, do
        # not re-schedule, it will be rescheduled when the other job
        # finishes
        if self.run_set.filter(return_dts__isnull=True).count():
            return

        if (self.reschedule_type and self.reschedule_interval_type
                and self.reschedule_interval):
            last_run = self.run_set.filter(is_manual=False)[0]

            if last_run.return_dts:
                if self.reschedule_type == 'AFTER_SCHEDULE_DTS':
                    reference_date = last_run.schedule_dts
                elif self.reschedule_type == 'AFTER_COMPLETE_DTS':
                    reference_date = last_run.return_dts

                try:
                    reschedule_date = self._get_reschedule_date(reference_date)

                    Run.objects.create(
                        job=self,
                        schedule_dts=reschedule_date,
                    )

                except RescheduleException:
                    t = get_template('job_runner/email/reschedule_failed.txt')
                    c = Context({
                        'job': self,
                        'hostname': settings.HOSTNAME,
                    })
                    email_body = t.render(c)

                    addresses = copy.copy(settings.JOB_RUNNER_ADMIN_EMAILS)
                    addresses.extend(self.get_notification_addresses())

                    if addresses:
                        send_mail(
                            'Reschedule error for: {0}'.format(self.title),
                            email_body,
                            settings.DEFAULT_FROM_EMAIL,
                            addresses
                        )

    def schedule_now(self):
        """
        Schedule the job to run now.
        """
        Run.objects.create(
            job=self,
            schedule_dts=timezone.now(),
        )

    def save(self, *args, **kwargs):
        t = Template(self.job_template.body)
        c = Context({'content': self.script_content_partial})
        self.script_content = t.render(c)
        super(Job, self).save(*args, **kwargs)

    def get_notification_addresses(self):
        """
        Return a ``list`` of notification addresses.
        """
        addresses = self.notification_addresses.strip().split('\n')
        addresses = [x.strip() for x in addresses if x.strip() != '']
        addresses.extend(self.job_template.get_notification_addresses())
        return addresses

    def _get_reschedule_incremented_dts(self, increment_date):
        """
        Increment the given ``reference_date`` with the reschedule interval.

        :param increment_date:
            An instance of :class:`!datetime.datetime` to increment.

        :return:
            An instance of :class:`!datetime.datetime`.

        """
        if self.reschedule_interval_type == 'MINUTE':
            return increment_date + timedelta(minutes=self.reschedule_interval)
        elif self.reschedule_interval_type == 'HOUR':
            return increment_date + timedelta(hours=self.reschedule_interval)
        elif self.reschedule_interval_type == 'DAY':
            return increment_date + timedelta(days=self.reschedule_interval)
        elif self.reschedule_interval_type == 'MONTH':
            # increment the dts with the number of days that are in the
            # ``reschedule_date``.
            reschedule_date = increment_date
            for x in range(self.reschedule_interval):
                days_in_month = calendar.monthrange(
                    reschedule_date.year, reschedule_date.month)
                reschedule_date = reschedule_date + timedelta(
                    days=days_in_month[1])
            return reschedule_date

    def _get_reschedule_date(
            self,
            reference_date,
            increment_date=None):
        """
        Return a reschedule datetime.

        This will take the reschedule exclude times into account when
        generating a new date/time.

        :param reference_date:
            The reference :class:`datetime.datetime` to generate the
            reschedule date from.

        :param increment_date:
            The :class:`datetime.datetime` to increment for calculating the
            reschedule date. This is optional and when not set, the
            ``reference_date`` will be used.

        :raises:
            :exc:`.RescheduleException` when the reschedule date can not
            be calculated. This is for example the case when a reschedule
            exclude exists which covers the whole day.

        :return:
            An instance of :class:`datetime.datetime`.

        """
        # make sure the reference_date is in the current tz, because we are
        # going to compare hours / minutes!
        reference_date = reference_date.astimezone(
            timezone.get_default_timezone())

        if not increment_date:
            increment_date = reference_date

        reschedule_date = self._get_reschedule_incremented_dts(increment_date)

        while reschedule_date < timezone.now():
            reschedule_date = self._get_reschedule_incremented_dts(
                reschedule_date)

        if (increment_date - reference_date) > timedelta(days=1):
            raise RescheduleException(
                'Unable to reschedule due to reschedule excludes')

        for exclude in self.rescheduleexclude_set.all():
            if (reschedule_date.time() >= exclude.start_time and
                    reschedule_date.time() <= exclude.end_time):
                return self._get_reschedule_date(
                    reference_date, reschedule_date)

        return reschedule_date


class RescheduleExclude(models.Model):
    """
    Exclude rules for rescheduling.
    """
    job = models.ForeignKey(Job)
    start_time = models.TimeField()
    end_time = models.TimeField()
    note = models.CharField(max_length=255, blank=True)

    def __unicode__(self):
        return u'{0} - {1}'.format(self.start_time, self.end_time)


class Run(models.Model):
    """
    Contains the data related to a (scheduled) job run.
    """
    job = models.ForeignKey(Job)
    schedule_dts = models.DateTimeField(db_index=True)
    enqueue_dts = models.DateTimeField(null=True, db_index=True)
    start_dts = models.DateTimeField(null=True, db_index=True)
    return_dts = models.DateTimeField(null=True, db_index=True)
    return_success = models.NullBooleanField(
        default=None, null=True, db_index=True)
    pid = models.PositiveIntegerField(null=True, default=None)
    is_manual = models.BooleanField(
        default=False, editable=False, db_index=True)
    schedule_children = models.BooleanField(
        default=True, editable=False, db_index=True)

    objects = RunManager()

    class Meta:
        ordering = (
            '-return_dts', '-start_dts', '-enqueue_dts', 'schedule_dts')

    @models.permalink
    def get_absolute_url(self):
        return ('job_runner:job_run', (), {
            'project_id': self.job.job_template.worker.project.pk,
            'job_id': self.job.pk,
            'run_id': self.pk,
        })

    def send_error_notification(self):
        """
        Send out an error notification e-mail.
        """
        t = get_template('job_runner/email/job_failed.txt')
        c = Context({
            'time_zone': settings.TIME_ZONE,
            'run': self,
            'hostname': settings.HOSTNAME,
        })
        email_body = t.render(c)

        addresses = copy.copy(settings.JOB_RUNNER_ADMIN_EMAILS)
        addresses.extend(self.job.get_notification_addresses())

        if addresses:
            send_mail(
                'Run error for: {0}'.format(self.job.title),
                email_body,
                settings.DEFAULT_FROM_EMAIL,
                addresses
            )


class KillRequest(models.Model):
    """
    Contains requests to kill active runs.
    """
    run = models.ForeignKey(Run)
    schedule_dts = models.DateTimeField(auto_now_add=True, db_index=True)
    enqueue_dts = models.DateTimeField(null=True, db_index=True)
    execute_dts = models.DateTimeField(null=True, db_index=True)

    objects = KillRequestManager()


class RunLog(models.Model):
    """
    Contains log output for runs.
    """
    run = models.OneToOneField(Run, related_name='run_log')
    content = models.TextField(null=True, default=None)

    class Meta:
        ordering = ('-run',)



from setuptools import find_packages, setup

import job_runner


setup(
    name='job-runner',
    version=job_runner.__version__,
    url='http://github.com/spilgames/jobrunner',
    author='Orne Brocaar',
    author_email='datawarehouse@spilgames.com',
    description='Tool for scheduling jobs and realtime monitoring',
    long_description=open('README.rst').read(),
    packages=find_packages(),
    include_package_data=True,
    scripts=['manage.py'],
    install_requires=[
        'django==1.4.2',
        'django-grappelli',
        'django_compressor',
        'south',
        'pytz',
        'pyzmq',
        'django-tastypie==0.9.12-alpha',
        'mimeparse==0.1.3',
        'python-dateutil==1.5',
    ],
    dependency_links=[
        (
            'https://github.com/toastdriven/django-tastypie/archive/'
            '27d17b7db7afd6c81da24f64f5b4562b59134582.tar.gz'
            '#egg=django-tastypie-0.9.12-alpha'
        ),
    ]
)



from django.test import TestCase
from django.utils import timezone
from mock import Mock, call

from job_runner.apps.job_runner.management.commands.broadcast_queue import (
    Command)
from job_runner.apps.job_runner.models import (
    Job, JobTemplate, KillRequest, Project, Run, Worker)


class CommandTestCase(TestCase):
    """
    Tests for :class:`.Command`.
    """
    fixtures = [
        'test_auth',
        'test_project',
        'test_worker',
        'test_job_template',
        'test_job',
    ]

    def test__broadcast_kill_requests(self):
        """
        Test :meth:`.Command._broadcast_kill_requests`.
        """
        run = Run.objects.get(pk=1)
        run.start_dts = timezone.now()
        run.pid = 1234
        run.save()
        KillRequest.objects.create(
            run=run,
            schedule_dts=timezone.now()
        )

        command = Command()

        publisher = Mock()
        command._broadcast_kill_requests(publisher)

        self.assertEqual([
            call([
                'master.broadcast.worker1',
                '{"action": "kill", "kill_request_id": 1}'
            ])
        ], publisher.send_multipart.call_args_list)

    def test__broadcast_runs(self):
        """
        Test :meth:`.Command._broadcast_runs`.
        """
        command = Command()

        publisher = Mock()
        command._broadcast_runs(publisher)

        self.assertEqual([
            call([
                'master.broadcast.worker1',
                '{"action": "enqueue", "run_id": 1}'
            ]),
            call([
                'master.broadcast.worker2',
                '{"action": "enqueue", "run_id": 2}'
            ]),
        ], publisher.send_multipart.call_args_list)

    def test__broadcast_runs_project_disabled_enqueue(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        for :class:`.Project`.
        """
        Project.objects.update(enqueue_is_enabled=False)
        command = Command()

        publisher = Mock()
        command._broadcast_runs(publisher)

        self.assertEqual([
        ], publisher.send_multipart.call_args_list)

    def test__broadcast_runs_worker_disabled_enqueue(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        for :class:`.Worker`.
        """
        Worker.objects.update(enqueue_is_enabled=False)
        command = Command()

        publisher = Mock()
        command._broadcast_runs(publisher)

        self.assertEqual([
        ], publisher.send_multipart.call_args_list)

    def test__broadcast_runs_job_template_disabled_enqueue(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        for :class:`.JobTemplate`.
        """
        JobTemplate.objects.update(enqueue_is_enabled=False)
        command = Command()

        publisher = Mock()
        command._broadcast_runs(publisher)

        self.assertEqual([
        ], publisher.send_multipart.call_args_list)

    def test__broadcast_runs_disabled_enqueue(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        for :class:`.Run`.
        """
        Job.objects.update(enqueue_is_enabled=False)
        command = Command()

        publisher = Mock()
        command._broadcast_runs(publisher)

        self.assertEqual([
        ], publisher.send_multipart.call_args_list)

    def test__broadcast_runs_with_active_run(self):
        """
        Test :meth:`.Command._broadcast_runs` with still one active run.
        """
        Run.objects.create(
            job=Job.objects.get(pk=1),
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
            start_dts=timezone.now(),
        )
        Run.objects.create(
            job=Job.objects.get(pk=2),
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
        )

        command = Command()
        publisher = Mock()
        command._broadcast_runs(publisher)

        self.assertEqual([], publisher.send_multipart.call_args_list)

    def test__broadcast_runs_disabled_enqueue_with_manual(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        but with manual one.
        """
        Job.objects.update(enqueue_is_enabled=False)
        Run.objects.filter(pk=2).update(is_manual=True)
        command = Command()

        publisher = Mock()
        command._broadcast_runs(publisher)

        self.assertEqual([
            call([
                'master.broadcast.worker2',
                '{"action": "enqueue", "run_id": 2}'
            ]),
        ], publisher.send_multipart.call_args_list)

    def test__broadcast_worker_ping(self):
        """
        Test :meth:`.Command._broadcast_worker_ping`.
        """
        command = Command()
        publisher = Mock()

        command._broadcast_worker_ping(publisher)

        self.assertEqual([
            call([
                'master.broadcast.worker1',
                '{"action": "ping"}'
            ]),
            call([
                'master.broadcast.worker2',
                '{"action": "ping"}'
            ]),
        ], publisher.send_multipart.call_args_list)



import os

from job_runner.settings.base import *


DEBUG = True
TEMPLATE_DEBUG = True


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(PROJECT_ROOT, '..', 'database.sqlite'),
    }
}


JOB_RUNNER_WS_SERVER = 'ws://localhost:5000/'


EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'


LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': '%(levelname)s - %(asctime)s - %(name)s: %(message)s',
        }
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'formatter': 'simple',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'DEBUG',
        },
    }
}


HOSTNAME = 'localhost:8000'



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Deleting field 'Run.return_log'
        db.delete_column('job_runner_run', 'return_log')


    def backwards(self, orm):
        # Adding field 'Run.return_log'
        db.add_column('job_runner_run', 'return_log',
                      self.gf('django.db.models.fields.TextField')(default=None, null=True),
                      keep_default=False)


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'job_runner.job': {
            'Meta': {'ordering': "('title',)", 'unique_together': "(('title', 'job_template'),)", 'object_name': 'Job'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'disable_enqueue_after_fails': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'enqueue_is_enabled': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'}),
            'fail_times': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'job_template': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.JobTemplate']"}),
            'notification_addresses': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'children'", 'null': 'True', 'to': "orm['job_runner.Job']"}),
            'reschedule_interval': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'reschedule_interval_type': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '6', 'blank': 'True'}),
            'reschedule_type': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '18', 'blank': 'True'}),
            'script_content': ('django.db.models.fields.TextField', [], {}),
            'script_content_partial': ('django.db.models.fields.TextField', [], {}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'job_runner.jobtemplate': {
            'Meta': {'ordering': "('title',)", 'object_name': 'JobTemplate'},
            'auth_groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'body': ('django.db.models.fields.TextField', [], {}),
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'enqueue_is_enabled': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'notification_addresses': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'worker': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.Worker']"})
        },
        'job_runner.killrequest': {
            'Meta': {'object_name': 'KillRequest'},
            'enqueue_dts': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'db_index': 'True'}),
            'execute_dts': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'db_index': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'run': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.Run']"}),
            'schedule_dts': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'})
        },
        'job_runner.project': {
            'Meta': {'ordering': "('title',)", 'object_name': 'Project'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'enqueue_is_enabled': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'notification_addresses': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'job_runner.rescheduleexclude': {
            'Meta': {'object_name': 'RescheduleExclude'},
            'end_time': ('django.db.models.fields.TimeField', [], {}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'job': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.Job']"}),
            'note': ('django.db.models.fields.CharField', [], {'max_length': '255', 'blank': 'True'}),
            'start_time': ('django.db.models.fields.TimeField', [], {})
        },
        'job_runner.run': {
            'Meta': {'ordering': "('-return_dts', '-start_dts', '-enqueue_dts', 'schedule_dts')", 'object_name': 'Run'},
            'enqueue_dts': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'db_index': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_manual': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            'job': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.Job']"}),
            'pid': ('django.db.models.fields.PositiveIntegerField', [], {'default': 'None', 'null': 'True'}),
            'return_dts': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'db_index': 'True'}),
            'return_success': ('django.db.models.fields.NullBooleanField', [], {'default': 'None', 'null': 'True', 'db_index': 'True', 'blank': 'True'}),
            'schedule_children': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'}),
            'schedule_dts': ('django.db.models.fields.DateTimeField', [], {'db_index': 'True'}),
            'start_dts': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'db_index': 'True'})
        },
        'job_runner.runlog': {
            'Meta': {'object_name': 'RunLog'},
            'content': ('django.db.models.fields.TextField', [], {'default': 'None', 'null': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'run': ('django.db.models.fields.related.OneToOneField', [], {'to': "orm['job_runner.Run']", 'unique': 'True'})
        },
        'job_runner.worker': {
            'Meta': {'ordering': "('title',)", 'object_name': 'Worker'},
            'api_key': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255', 'db_index': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'enqueue_is_enabled': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'notification_addresses': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'project': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['job_runner.Project']"}),
            'secret': ('django.db.models.fields.CharField', [], {'max_length': '255', 'db_index': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        }
    }

    complete_apps = ['job_runner']